/*
//   1D Solid Mechanics for Truss Elements solved by
//   CG (Conjugate Gradient) Method using 2nd-order Elements
//
//   d/dx(EdU/dx) + F = 0
//   U=0@x=0  
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

int main(){
        int NE, N, NPLU, IterMax, errno, NPLU0;
	int R, Z, Q, P, DD;

	double dX, Resid, Eps, Area, F, Young;
	double X1, X2, X3, U1, U2, U3, DL, Strain, Sigma, Ck;
	double *U, *Rhs, *X;
	double *Diag, *AMat;
	double **W;

	int *Index, *Item, *Icelnod;
        double Kmat[3][3], Emat[3][3];

	int i, j, in1, in2, in3, k, icel, k12, k13, k21, k23, k31, k32, jS;
	int iter;
	FILE *fp;
	double BNorm2, Rho, Rho1=0.0, C1, Alpha, DNorm2;
	int ierr = 1;

/*
// +-------+
// | INIT. |
// +-------+
*/
	fp = fopen("input2.dat", "r");
	assert(fp != NULL);
	fscanf(fp, "%d", &NE);
	fscanf(fp, "%lf %lf %lf %lf", &dX, &F, &Area, &Young);
	fscanf(fp, "%d", &IterMax);
	fscanf(fp, "%lf", &Eps);
	fclose(fp);

        N    = 2*NE + 1;
        NPLU0= 2*2 + NE*2 + (NE-1)*4;

	U    = calloc(N, sizeof(double));
	X    = calloc(N, sizeof(double));
	Diag = calloc(N, sizeof(double));
	AMat = calloc(NPLU0, sizeof(double));
	Rhs = calloc(N, sizeof(double));
	Index= calloc(N+1, sizeof(int));
	Item = calloc(NPLU0, sizeof(int));
	Icelnod= calloc(3*NE, sizeof(int));

        W = (double **)malloc(sizeof(double *)*4);
        if(W == NULL) {
                fprintf(stderr, "Error: %s\n", strerror(errno));
                return -1;
        }
        for(i=0; i<4; i++) {
                W[i] = (double *)malloc(sizeof(double)*N);
                if(W[i] == NULL) {
                        fprintf(stderr, "Error: %s\n", strerror(errno));
                        return -1;
                }
        }

	for(i=0;i<N;i++)    U[i] = 0.0;
	for(i=0;i<N;i++) Diag[i] = 0.0;
	for(i=0;i<N;i++)  Rhs[i] = 0.0;
	for(k=0;k<NPLU0;k++)  AMat[k] = 0.0;
	for(i=0;i<N;i++) X[i]= i*dX*0.5;
	for(icel=0;icel<NE;icel++){
		Icelnod[3*icel  ]= 2*icel;     
		Icelnod[3*icel+1]= 2*icel+1;     
		Icelnod[3*icel+2]= 2*icel+2;     
	}

        Kmat[0][0]= +14.0/6.;
        Kmat[0][1]= -16.0/6.;
        Kmat[0][2]=  +2.0/6.;
        Kmat[1][0]= -16.0/6.;
        Kmat[1][1]= +32.0/6.;
        Kmat[1][2]= -16.0/6.;
        Kmat[2][0]=  +2.0/6.;
        Kmat[2][1]= -16.0/6.;
        Kmat[2][2]= +14.0/6.;
/*
// +--------------+
// | CONNECTIVITY |
// +--------------+
*/
	Index[0]= 0;
	for(i=1;i<N;i++) {
	  if (i%2==1) {Index[i]=4;
	  }else       {Index[i]=2;}}

	Index[1]= 2;
	Index[N]= 2;


	for(i=0;i<N;i++){
		Index[i+1]= Index[i+1] + Index[i];     
	}

        NPLU= Index[N];

	for(i=0;i<N;i++){
		int jS = Index[i];
		if(i == 0){
			Item[jS  ] = i+1;
			Item[jS+1] = i+2;
		}else if(i == N-1){
			Item[jS  ] = i-2;
			Item[jS+1] = i-1;
		}else{
		  if (i%2==1){
			Item[jS  ] = i-1;
			Item[jS+1] = i+1;
		  } else       {
			Item[jS  ] = i-2;
			Item[jS+1] = i-1;
			Item[jS+2] = i+1;
			Item[jS+3] = i+2;}
		}
	}


/*
// +-----------------+
// | MATRIX assemble |
// +-----------------+
*/
	for(icel=0;icel<NE;icel++){
	  in1= Icelnod[3*icel];
	  in2= Icelnod[3*icel+1];
	  in3= Icelnod[3*icel+2];
	  X1 = X[in1];
	  X2 = X[in2];
	  X3 = X[in3];
          DL = fabs(X3-X1);

          Ck= Area*Young/DL;
          Emat[0][0]= Ck*Kmat[0][0];
          Emat[0][1]= Ck*Kmat[0][1];
          Emat[0][2]= Ck*Kmat[0][2];
          Emat[1][0]= Ck*Kmat[1][0];
          Emat[1][1]= Ck*Kmat[1][1];
          Emat[1][2]= Ck*Kmat[1][2];
          Emat[2][0]= Ck*Kmat[2][0];
          Emat[2][1]= Ck*Kmat[2][1];
          Emat[2][2]= Ck*Kmat[2][2];

          Diag[in1]= Diag[in1] + Emat[0][0];
          Diag[in2]= Diag[in2] + Emat[1][1];
          Diag[in3]= Diag[in3] + Emat[2][2];

          if (icel==0){k12=Index[in1];
                       k13=Index[in1]+1;
	        }else {k12=Index[in1]+2;
	               k13=Index[in1]+3;}

	    k21=Index[in2];
	    k23=Index[in2]+1;

	    k31=Index[in3];
	    k32=Index[in3]+1;

	    AMat[k12]= AMat[k12] + Emat[0][1];
	    AMat[k13]= AMat[k13] + Emat[0][2];
	    AMat[k21]= AMat[k21] + Emat[1][0];
	    AMat[k23]= AMat[k23] + Emat[1][2];
	    AMat[k31]= AMat[k31] + Emat[2][0];
	    AMat[k32]= AMat[k32] + Emat[2][1];
	}

/*
// +---------------------+
// | BOUNDARY conditions |
// +---------------------+
*/

/* X=Xmin */
        i=0;
        jS= Index[i];
        AMat[jS  ]= 0.0;
        AMat[jS+1]= 0.0;
        Diag[i ]= 1.0;
        Rhs [i ]= 0.0;

	for(k=0;k<NPLU;k++){
	  if(Item[k]==0){AMat[k]=0.0;
	  }}

/* X=Xmax */
        i=N-1;
        Rhs[i]= F;


/*
// +---------------+
// | CG iterations |
// +---------------+
*/
	R = 0;
	Z = 1;
	Q = 1;
	P = 2;
	DD= 3;

	for(i=0;i<N;i++){
		W[DD][i]= 1.0 / Diag[i];
	}

/*
//-- {r0}= {b} - [A]{xini} |
*/
	for(i=0;i<N;i++){
		W[R][i] = Diag[i]*U[i];
		for(j=Index[i];j<Index[i+1];j++){
			W[R][i] += AMat[j]*U[Item[j]];
		}
	}

	BNorm2 = 0.0;
	for(i=0;i<N;i++){
		BNorm2 += Rhs[i] * Rhs[i];
		W[R][i] = Rhs[i] - W[R][i];
	}


/* *******************************************************************
*/
	for(iter=1;iter<=IterMax;iter++){
/*
//-- {z}= [Minv]{r}
*/
		for(i=0;i<N;i++){
			W[Z][i] = W[DD][i] * W[R][i];
		}

/*
//-- RHO= {r}{z}
*/
		Rho= 0.0;
		for(i=0;i<N;i++){
			Rho += W[R][i] * W[Z][i];
		}


/*
//-- {p} = {z} if      ITER=1  
//   BETA= RHO / RHO1  otherwise 
*/
		if(iter == 1){
			for(i=0;i<N;i++){
				W[P][i] = W[Z][i];
			}
		}else{
			double Beta = Rho / Rho1;
			for(i=0;i<N;i++){
				W[P][i] = W[Z][i] + Beta*W[P][i];
			}
		}

/*
//-- {q}= [A]{p}
*/
		for(i=0;i<N;i++){
			W[Q][i] = Diag[i] * W[P][i];
			for(j=Index[i];j<Index[i+1];j++){
				W[Q][i] += AMat[j]*W[P][Item[j]];
			}
		}

/*
//-- ALPHA= RHO / {p}{q}
*/
		C1 = 0.0;
		for(i=0;i<N;i++){
			C1 += W[P][i] * W[Q][i];
		}
		Alpha = Rho / C1;


/*
//-- {x}= {x} + ALPHA*{p}
//   {r}= {r} - ALPHA*{q}
*/
		for(i=0;i<N;i++){
			U[i]    += Alpha * W[P][i];
			W[R][i] -= Alpha * W[Q][i];
		}

		DNorm2 = 0.0;
		for(i=0;i<N;i++){
			DNorm2 += W[R][i] * W[R][i];
		}

		Resid = sqrt(DNorm2/BNorm2);

          	  if((iter)%1000 == 0){
		  printf("%8d%s%16.6e\n",
			 iter, " iters, RESID=", Resid);
		  }

		if(Resid <= Eps){
			ierr = 0;
			break;
		}

		Rho1 = Rho;

	}
/* ********************************************************************
*/

/*
//-- OUTPUT
*/
        printf("%8d%s%16.6e\n", iter, " iters, RESID=", Resid);


        printf("\n%s\n", "### DISPLACEMENT");
	for(i=0;i<N;i++){
	  printf("%8d%16.6E%16.6E\n", i+1, U[i], X[i]*F/Area/Young);
	}

	return ierr;
	}
