!C
!C   1D Solid Mechanics for Truss Elements solved by
!C   CG (Conjugate Gradient) Method using 2nd-order Elements
!C
!C   d/dx(EdU/dx) + F = 0
!C   U=0@x=0  
!C
      program solid1D
      implicit REAL*8 (A-H,O-Z)

      integer :: N, NPLU, ITERmax
      integer :: R, Z, P, Q, DD

      real(kind=8) :: dX, RESID, EPS 
      real(kind=8) :: AREA, F, YOUNG
      real(kind=8), dimension(:), allocatable :: U, RHS, X
      real(kind=8), dimension(:  ), allocatable :: DIAG, AMAT
      real(kind=8), dimension(:,:), allocatable :: W

      real(kind=8), dimension(3,3) :: KMAT, EMAT

      integer, dimension(:), allocatable :: ICELNOD
      integer, dimension(:), allocatable :: INDEX, ITEM

!C
!C +-------+
!C | INIT. |
!C +-------+
!C===
      open  (11, file='input2.dat', status='unknown')
       read (11,*) NE
       read (11,*) dX, F, AREA, YOUNG
       read (11,*) ITERmax
       read (11,*) EPS
      close (11)

      N    = 2*NE + 1
      NPLU0= 2*2 + NE*2 + (NE-1)*4

      allocate (U(N), DIAG(N), AMAT(NPLU0), RHS(N))
      allocate (ICELNOD(3*NE), X(N))
      allocate (INDEX(0:N), ITEM(NPLU0), W(N,4))
         U= 0.d0
      AMAT= 0.d0
      DIAG= 0.d0
       RHS= 0.d0
         X= 0.d0

      do i= 1, N
        X(i)= dfloat(i-1)*dX*0.50d0
      enddo

      do icel= 1, NE
        ICELNOD(3*icel-2)= 2*icel - 1
        ICELNOD(3*icel-1)= 2*icel
        ICELNOD(3*icel  )= 2*icel + 1
      enddo

      KMAT(1,1)= +14.d0/6.d0
      KMAT(1,2)= -16.d0/6.d0
      KMAT(1,3)=  +2.d0/6.d0
      KMAT(2,1)= -16.d0/6.d0
      KMAT(2,2)= +32.d0/6.d0
      KMAT(2,3)= -16.d0/6.d0
      KMAT(3,1)=  +2.d0/6.d0
      KMAT(3,2)= -16.d0/6.d0
      KMAT(3,3)= +14.d0/6.d0
!C===

!C
!C +--------------+
!C | CONNECTIVITY |
!C +--------------+
!C===
      INDEX(0)= 0
      do i= 1, N
        if (mod(i,2).eq.1) then
          INDEX(i)= 4
         else 
          INDEX(i)= 2
        endif
      enddo

      INDEX(1)= 2 
      INDEX(N)= 2

      do i= 1, N
        INDEX(i)= INDEX(i) + INDEX(i-1)      
      enddo

      NPLU= INDEX(N)

      do i= 1, N
        jS= INDEX(i-1)
        if (i.eq.1) then
          ITEM(jS+1)= i+1
          ITEM(jS+2)= i+2
         else if                                                        &
     &     (i.eq.N) then
          ITEM(jS+1)= i-2
          ITEM(jS+2)= i-1
         else
          if (mod(i,2).eq.1) then
            ITEM(jS+1)= i-2            
            ITEM(jS+2)= i-1
            ITEM(jS+3)= i+1
            ITEM(jS+4)= i+2
           else
            ITEM(jS+1)= i-1
            ITEM(jS+2)= i+1
          endif
        endif
      enddo
!C===

!C
!C +-----------------+
!C | MATRIX ASSEMBLE |
!C +-----------------+
!C===
      do icel= 1, NE
        in1= ICELNOD(3*icel-2)
        in2= ICELNOD(3*icel-1)
        in3= ICELNOD(3*icel  )
        X1 = X(in1)
        X2 = X(in2)
        X3 = X(in3)
        DL = dabs(X3-X1)

        cK= AREA*YOUNG/DL
        EMAT(1,1)= Ck*KMAT(1,1)
        EMAT(1,2)= Ck*KMAT(1,2)
        EMAT(1,3)= Ck*KMAT(1,3)
        EMAT(2,1)= Ck*KMAT(2,1)
        EMAT(2,2)= Ck*KMAT(2,2)
        EMAT(2,3)= Ck*KMAT(2,3)
        EMAT(3,1)= Ck*KMAT(3,1)
        EMAT(3,2)= Ck*KMAT(3,2)
        EMAT(3,3)= Ck*KMAT(3,3)

        DIAG(in1)= DIAG(in1) + EMAT(1,1)
        DIAG(in2)= DIAG(in2) + EMAT(2,2)
        DIAG(in3)= DIAG(in3) + EMAT(3,3)

        if (icel.eq.1) then
          k12= INDEX(in1-1) + 1  
          k13= INDEX(in1-1) + 2  
         else
          k12= INDEX(in1-1) + 3  
          k13= INDEX(in1-1) + 4  
        endif

        k21= INDEX(in2-1) + 1  
        k23= INDEX(in2-1) + 2

        k31= INDEX(in3-1) + 1  
        k32= INDEX(in3-1) + 2  

        AMAT(k12)= AMAT(k12) + EMAT(1,2)
        AMAT(k13)= AMAT(k13) + EMAT(1,3)
        AMAT(k21)= AMAT(k21) + EMAT(2,1)
        AMAT(k23)= AMAT(k23) + EMAT(2,3)
        AMAT(k31)= AMAT(k31) + EMAT(3,1)
        AMAT(k32)= AMAT(k32) + EMAT(3,2)
      enddo
!C===

!C
!C +---------------------+
!C | BOUNDARY CONDITIONS |
!C +---------------------+
!C===

!C
!C-- X=Xmin
      i= 1
      jS= INDEX(i-1)

      AMAT(jS+1)= 0.d0
      AMAT(jS+2)= 0.d0
      DIAG(i)= 1.d0
      RHS (i)= 0.d0
      do k= 1, NPLU
        if (ITEM(k).eq.1) AMAT(k)= 0.d0
      enddo

!C
!C-- X=Xmax
      i= N
      RHS(i)= F
!C===

!C
!C +---------------+
!C | CG iterations |
!C +---------------+
!C===
      R = 1
      Z = 2
      Q = 2
      P = 3
      DD= 4

      do i= 1, N
        W(i,DD)= 1.0D0 / DIAG(i)
      enddo

!C
!C-- {r0}= {b} - [A]{xini} |

      do i= 1, N
        W(i,R) = DIAG(i)*U(i)
        do j= INDEX(i-1)+1, INDEX(i)
          W(i,R) = W(i,R) + AMAT(j)*U(ITEM(j))
        enddo
      enddo

      BNRM2= 0.0D0
      do i= 1, N
        BNRM2 = BNRM2  + RHS(i) **2
        W(i,R)= RHS(i) - W(i,R)
      enddo

!C********************************************************************
      do iter= 1, ITERmax
!C
!C-- {z}= [Minv]{r}

      do i= 1, N
        W(i,Z)= W(i,DD) * W(i,R)
      enddo

!C
!C-- RHO= {r}{z}

      RHO= 0.d0
      do i= 1, N
        RHO= RHO + W(i,R)*W(i,Z)   
      enddo     

!C
!C-- {p} = {z} if      ITER=1  
!C   BETA= RHO / RHO1  otherwise 

      if ( iter.eq.1 ) then
        do i= 1, N
          W(i,P)= W(i,Z)
        enddo
       else
         BETA= RHO / RHO1
         do i= 1, N
           W(i,P)= W(i,Z) + BETA*W(i,P)
         enddo
      endif

!C
!C-- {q}= [A]{p}

      do i= 1, N
        W(i,Q) = DIAG(i)*W(i,P)
        do j= INDEX(i-1)+1, INDEX(i)
          W(i,Q) = W(i,Q) + AMAT(j)*W(ITEM(j),P)
        enddo
      enddo

!C
!C-- ALPHA= RHO / {p}{q}

      C1= 0.d0
      do i= 1, N
        C1= C1 + W(i,P)*W(i,Q)
      enddo
      ALPHA= RHO / C1

!C
!C-- {x}= {x} + ALPHA*{p}
!C   {r}= {r} - ALPHA*{q}

      do i= 1, N
        U(i)  = U(i)   + ALPHA * W(i,P)
        W(i,R)= W(i,R) - ALPHA * W(i,Q)
      enddo

      DNRM2 = 0.0
      do i= 1, N
        DNRM2= DNRM2 + W(i,R)**2
      enddo

        RESID= dsqrt(DNRM2/BNRM2)

        if (mod(iter,1000).eq.0) then
        write (*,'(i8, a, 1pe16.6, a, 1pe16.6)')                        &
     &         iter, ' iters, RESID=', RESID, ' U(N)= ', U(N)
        endif

        if ( RESID.le.EPS) goto 900
        RHO1 = RHO

      enddo
!C********************************************************************

      IER = 1

  900 continue
!C===

!C
!C-- OUTPUT
        write (*,'(i8, a, 1pe16.6, a, 1pe16.6)')                        &
     &         iter, ' iters, RESID=', RESID, ' U(N)= ', U(N)

      write (*,'(/a)') '### DISPLACEMENT'
      do i= 1, N
        write (*,'(i8, 2(1pe16.6))') i, U(i), X(i)*F/AREA/YOUNG
      enddo

      end program solid1D
